/*_############################################################################
  _##
  _##  SNMP4J - SnmpBuilder.java
  _##
  _##  Copyright (C) 2003-2021  Frank Fock (SNMP4J.org)
  _##
  _##  Licensed under the Apache License, Version 2.0 (the "License");
  _##  you may not use this file except in compliance with the License.
  _##  You may obtain a copy of the License at
  _##
  _##      http://www.apache.org/licenses/LICENSE-2.0
  _##
  _##  Unless required by applicable law or agreed to in writing, software
  _##  distributed under the License is distributed on an "AS IS" BASIS,
  _##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  _##  See the License for the specific language governing permissions and
  _##  limitations under the License.
  _##
  _##########################################################################*/
package org.snmp4j.fluent;

import org.snmp4j.MessageDispatcher;
import org.snmp4j.Snmp;
import org.snmp4j.TransportMapping;
import org.snmp4j.mp.CounterSupport;
import org.snmp4j.mp.MPv1;
import org.snmp4j.mp.MPv2c;
import org.snmp4j.mp.MPv3;
import org.snmp4j.security.SecurityModels;
import org.snmp4j.security.SecurityProtocols;
import org.snmp4j.security.TSM;
import org.snmp4j.security.USM;
import org.snmp4j.smi.*;
import org.snmp4j.transport.*;
import org.snmp4j.transport.tls.TlsTmSecurityCallback;
import org.snmp4j.util.MultiThreadedMessageDispatcher;
import org.snmp4j.util.ThreadPool;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.security.cert.X509Certificate;

/**
 * Builds a {@link Snmp} instance through step-by-step configuration.
 *
 * @author Frank Fock
 * @since 3.5.0
 */
public class SnmpBuilder {

    protected final Snmp snmp;
    protected final SecurityProtocols securityProtocols;
    protected final SecurityModels securityModels;
    protected CounterSupport counterSupport;
    protected ThreadPool multiThreadedDispatcherPool;
    protected String responderPoolName = "SnmpDispatcherPool";

    public SnmpBuilder() {
        this(new Snmp());
    }

    public SnmpBuilder(Snmp snmp) {
        this.snmp = snmp;
        this.snmp.getMessageDispatcher().addCommandResponder(snmp);
        counterSupport = CounterSupport.getInstance();
        securityProtocols = new SecurityProtocols(SecurityProtocols.SecurityProtocolSet.defaultSecurity);
        securityModels = new SecurityModels();
    }

    public <A extends Address> TargetBuilder<A> target(A address) {
        return TargetBuilder.forAddress(this, address);
    }

    public SnmpBuilder counterSupport(CounterSupport counterSupport) {
        this.counterSupport = counterSupport;
        return this;
    }

    public SnmpBuilder v1() {
        snmp.getMessageDispatcher().addMessageProcessingModel(new MPv1());
        return this;
    }

    public SnmpBuilder v2c() {
        snmp.getMessageDispatcher().addMessageProcessingModel(new MPv2c());
        return this;
    }

    public SnmpBuilder v3() {
        return v3(MPv3.createLocalEngineID());
    }

    /**
     * Adds a SNMPv3 message processing model with a random engine ID based on the supplied ID string.
     *
     * @param id
     *         an ID string suitable to build a local engine ID.
     * @return
     *    this builder
     */
    public SnmpBuilder v3(OctetString id) {
        return v3(MPv3.createLocalEngineID(id));
    }

    public SnmpBuilder v3(byte[] localEngineID) {
        MPv3 mpv3 = new MPv3(localEngineID);
        mpv3.setSecurityModels(securityModels);
        mpv3.setSecurityProtocols(securityProtocols);
        snmp.getMessageDispatcher().addMessageProcessingModel(mpv3);
        return this;
    }

    public SnmpBuilder threads(int numThreads) {
        MessageDispatcher messageDispatcher = snmp.getMessageDispatcher();
        if (messageDispatcher instanceof MultiThreadedMessageDispatcher) {
            messageDispatcher = ((MultiThreadedMessageDispatcher)messageDispatcher).getDispatcher();
        }
        if (multiThreadedDispatcherPool != null) {
            multiThreadedDispatcherPool.cancel();
        }
        this.multiThreadedDispatcherPool = ThreadPool.create(responderPoolName, numThreads);
        snmp.setMessageDispatcher(new MultiThreadedMessageDispatcher(
                this.multiThreadedDispatcherPool, messageDispatcher));
        return this;
    }

    /**
     * Creates a USM for the specified engine ID and engine boots counter. This method requires that {@link #v3(byte[])}
     * has been already called before, otherwise a {@link NullPointerException} will be thrown.
     * @return
     *    this builder.
     */
    public SnmpBuilder usm() {
        return usm(new OctetString(((MPv3)
                snmp.getMessageDispatcher().getMessageProcessingModel(MPv3.ID)).getLocalEngineID()), 0);
    }

    /**
     * Creates a USM for the specified engine ID and engine boots counter.
     * @param localEngineID
     *         the local engine ID.
     * @param engineBoots
     *         the number of engine boots.
     * @return
     *    this builder.
     */
    public SnmpBuilder usm(OctetString localEngineID, int engineBoots) {
        securityModels.addSecurityModel(new USM(securityProtocols, localEngineID, engineBoots));
        return this;
    }

    /**
     * Creates a Transport Security Model as defined by RFC 5591 and adds it to the built {@link Snmp} instance.
     * @param localEngineID
     *    the engine ID of the SNMP entity using this transport security model. The local engine ID must be globally
     *    unique.
     * @param usePrefix
     *    the snmpTsmConfigurationUsePrefix flag as defined in RFC 5591.
     * @return
     *    this builder.
     */
    public SnmpBuilder tsm(OctetString localEngineID, boolean usePrefix) {
        securityModels.addSecurityModel(new TSM(localEngineID, usePrefix));
        return this;
    }

    public SnmpBuilder udp() throws IOException {
        TransportMapping<UdpAddress> tm = new DefaultUdpTransportMapping(new UdpAddress(new InetSocketAddress(0).getAddress(), 0));
        snmp.getMessageDispatcher().addTransportMapping(tm);
        tm.addTransportListener(snmp.getMessageDispatcher());
        return this;
    }

    public SnmpBuilder udp(UdpAddress... listenAddresses) throws IOException {
        for (UdpAddress listenAddress : listenAddresses) {
            UdpTransportMapping udpTransportMapping = new DefaultUdpTransportMapping(listenAddress);
            snmp.getMessageDispatcher().addTransportMapping(udpTransportMapping);
            udpTransportMapping.addTransportListener(snmp.getMessageDispatcher());
        }
        return this;
    }

    public SnmpBuilder tcp() throws IOException {
        TransportMapping<TcpAddress> tm = new DefaultTcpTransportMapping(new TcpAddress(new InetSocketAddress(0).getAddress(), 0));
        snmp.getMessageDispatcher().addTransportMapping(tm);
        tm.addTransportListener(snmp.getMessageDispatcher());
        return this;
    }

    public SnmpBuilder tcp(TcpAddress... listenAddresses) throws IOException {
        for (TcpAddress listenAddress : listenAddresses) {
            TcpTransportMapping<?> tcpTransportMapping = new DefaultTcpTransportMapping(listenAddress);
            snmp.getMessageDispatcher().addTransportMapping(tcpTransportMapping);
            tcpTransportMapping.addTransportListener(snmp.getMessageDispatcher());
        }
        return this;
    }

    public SnmpBuilder dtls() throws IOException {
        DTLSTM dtlstm = new DTLSTM(new DtlsAddress(new InetSocketAddress(0).getAddress(), 0), false);
        snmp.getMessageDispatcher().addTransportMapping(dtlstm);
        dtlstm.addTransportListener(snmp.getMessageDispatcher());
        return this;
    }

    public SnmpBuilder dtls(TlsTmSecurityCallback<X509Certificate> securityCallback,
                            DtlsAddress... listenAddresses) throws IOException {
        return dtls(securityCallback, null, listenAddresses);
    }

    public SnmpBuilder dtls(TlsTmSecurityCallback<X509Certificate> securityCallback, String[] dtlsProtocolVersions,
                            DtlsAddress... listenAddresses) throws IOException {
        for (DtlsAddress listenAddress : listenAddresses) {
            DTLSTM dtlsTransportMapping = new DTLSTM(securityCallback, listenAddress, counterSupport, true);
            if (dtlsProtocolVersions != null && dtlsProtocolVersions.length > 0) {
                dtlsTransportMapping.setProtocolVersions(dtlsProtocolVersions);
            }
            snmp.getMessageDispatcher().addTransportMapping(dtlsTransportMapping);
            dtlsTransportMapping.addTransportListener(snmp.getMessageDispatcher());
        }
        return this;
    }

    public SnmpBuilder tls() throws IOException {
        TLSTM tlstm = new TLSTM(new TlsAddress(new InetSocketAddress(0).getAddress(), 0), false);
        snmp.getMessageDispatcher().addTransportMapping(tlstm);
        tlstm.addTransportListener(snmp.getMessageDispatcher());
        return this;
    }

    public SnmpBuilder tls(TlsTmSecurityCallback<X509Certificate> securityCallback,
                           TlsAddress... listenAddresses) throws IOException {
        return tls(securityCallback, null, listenAddresses);
    }

    public SnmpBuilder tls(TlsTmSecurityCallback<X509Certificate> securityCallback, String[] tlsProtocolVersions,
                           TlsAddress... listenAddresses) throws IOException {
        for (TlsAddress listenAddress : listenAddresses) {
            TLSTM tlsTransportMapping = new TLSTM(securityCallback, listenAddress, counterSupport, true);
            if (tlsProtocolVersions != null && tlsProtocolVersions.length > 0) {
                tlsTransportMapping.setProtocolVersions(tlsProtocolVersions);
            }
            snmp.getMessageDispatcher().addTransportMapping(tlsTransportMapping);
            tlsTransportMapping.addTransportListener(snmp.getMessageDispatcher());
        }
        return this;
    }

    public SnmpBuilder securityProtocols(SecurityProtocols.SecurityProtocolSet securityProtocolSet) {
        this.securityProtocols.removeAll();
        this.securityProtocols.addPredefinedProtocolSet(securityProtocolSet);
        return this;
    }

    public Snmp build() throws IOException {
        snmp.listen();
        return snmp;
    }
}
